"""
Copyright (c) 2024 VMware, Inc. All Rights Reserved.
"""

import json
import os
import subprocess

from idemgrpc import messages_pb2


def execute_script_task_helper(hub, task):
    
    if not task.executionContext:
        err = "Task has no executionContext"
        hub.log.error(err)
        return 1, err
    
    task_id = task.taskIdentity.taskId
    org_id = task.taskIdentity.identity.orgId
    project_id = task.taskIdentity.identity.projectId
    project_type = task.taskIdentity.identity.projectType
    executor_path = os.getenv("SCRIPT_TASK_DIRECTORY") + '/executor.sh'
    
    execution_context = json.loads(task.executionContext)
    cmd = execution_context.get("cmd")
    
    env = {
        "REQUESTOR": task.requester,
        "TASK_ID": task_id,
        "ORG_ID": org_id,
        "PROJECT_ID": project_id,
    }
    env = {**env, **os.environ.copy()}

    entity_id = execution_context.get("entityId")
    if entity_id is None or not entity_id:
        env["ENTITY_ID"] = ""
    else:
        env["ENTITY_ID"] = entity_id

    if hub.idemd.worker.workerType == messages_pb2.WorkerIdentity.WorkerType.REMOTE:
        env["IS_REMOTE_WORKER"] = "true"
    else:
        env["IS_REMOTE_WORKER"] = "false"

    args = [executor_path, cmd]

    args_in = execution_context.get("params")
    if args_in and isinstance(args_in, dict):
        for k, v in args_in.items():
            args.extend([k, v])
    
    env_in = execution_context.get("environment")
    if env_in and isinstance(env_in, dict):
        env = {**env, **env_in}
    
    proc_timeout = (
        task.expiresInSeconds
        if task.expiresInSeconds is not None and task.expiresInSeconds > 0
        else hub.idemd.global_task_timeout
    )
   
    message, proc = None, None
    try:
        hub.log.debug(f"About to execute script task for org: {org_id}, project {project_id}, task: {task_id}")
        proc = subprocess.Popen(
            args=args,
            env=env,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        try:
            stdout, stderr = proc.communicate(timeout=proc_timeout)
            hub.log.info(
                f" Process ID {proc.pid} for task ID {task.taskIdentity.taskId}"
            )
            if stdout:
                message = stdout.decode()
                hub.log.info(f"[stdout]\n{message}")
            if stderr:
                message = stderr.decode()
                hub.log.error(f"[stderr]\n{message}")
            return_code = proc.returncode
        except subprocess.TimeoutExpired:
            hub.log.error(
                f"Timeout occurred for task"
                f" {task.taskIdentity.taskId}. Killing the process with PID "
                f"{proc.pid} and will proceed with idem service update"
            )
            try:
                # On timeout error, explicitly kill the process
                proc.kill()
            except OSError:
                # Ignore 'no such process' error
                pass
            # exit code for timeout
            return_code = 124
            message = f"Timeout occurred for task {task.taskIdentity.taskId}"
        except Exception as exc:
            # Any exception other than the TimeoutError will be caught here. This can happen when the process
            # itself fails with some Exception before timeout.
            # await proc.communicate()
            hub.log.error(
                f"Exception Occurred. Task {task.taskIdentity.taskId} failed with: {exc}"
            )
            message = repr(exc)
            return_code = 1
    except Exception as err:
        hub.log.error(
            f"Error occurred when spawning a new process for task {task.taskIdentity.taskId} - {str(err)}"
        )
        return_code = 1
        message = repr(err)
    
    return return_code, message

import uuid

# This is just test function will be removed once code stablizes
def tester():
    os.environ['SCRIPT_DIRECTORY'] = '/home/ncd/vmware/njakhar/bellevue-ci/idem-service-worker/scripts'
    os.environ['SCRIPT_TASK_DIRECTORY'] = '/home/ncd/vmware/njakhar/bellevue-ci/idem-service-worker/script-task'
    os.environ['GIT_TYPE'] = 'gitlab'
    os.environ['GIT_ENDPOINT'] = 'https://gitlab.eng.vmware.com'
    os.environ['GIT_PERSONAL_TOKEN'] = 's6i7SMT7RLDc4DfP1D5x'
    os.environ['LEMANS_KEY'] = 'acc75ldO6msl6vsNIBifYzZutDWQ8uZh'
    os.environ['LEMANS_BASE_URL'] = 'https://data.be.symphony-dev.com'
    os.environ['WORKER_NAME'] = str(uuid.uuid4())

    repo_collection_context = '{ "cmd": "rw-repo-collection-task" }'
    sbom_generation_context = '{ "cmd": "rw-repo-sbom-task", "params": { "-p": "cmbu/guardrails-service", "-b": "master", "-c": "473ce8d71bda4e5ed18779f5522cde222ffeb216" } }'
    sbom_generation_context_latest = '{ "cmd": "rw-repo-sbom-task", "params": { "-p": "cmbu/guardrails-service", "-b": "master" } }'

    execute_script_task_helper(type("", (dict,), {
            'log': type("", (dict,), {'info': print, 'debug': print, 'error': print}),
            'idemd': type("", (dict,), {'worker': type("", (dict,), {'workerType': 1})})
        }), type(
            "",
            (dict,), {
                'expiresInSeconds': 10000,
                'executionContext': sbom_generation_context_latest,
                'requester': 'ASSESSMENT',
                'taskIdentity': type(
                    "",
                    (dict,), {
                        'taskId': str(uuid.uuid4()),
                        'identity': type(
             "",
                            (dict,), {
                                'orgId': str(uuid.uuid4()),
                                'projectId': 'default',
                                'projectType': 'dummy',
                            },
                        ),
                    },
                ),
            }
        ),
    )

#tester()
